==> UDFclient.0.3.3.tgz: UDFclient.0.3.3/scsi_subr.c <==
/*	$NetBSD: scsi_subr.c,v 1.7 2002/10/08 20:17:06 soren Exp $	*/

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Charles M. Hannum; Jason R. Thorpe of the Numerical Aerospace
 * Simulation Facility, NASA Ames Research Center.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Small changes, generalisations and Linux support by Reinoud Zandijk
 * <reinoud@netbsd.org>.
 *
 */


/*
 * SCSI support subroutines.
 *
 */

#include <sys/param.h>
#include <sys/ioctl.h>
#include <err.h> 
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <inttypes.h>
#include <assert.h>

#include "scsi_sense.h"
#include "scsilib.h"


#define	STRVIS_ISWHITE(x) ((x) == ' ' || (x) == '\0' || (x) == (u_char)'\377')
int scsilib_verbose = 0;


#ifdef UDF_SCSIPI
	/*
	 * scsipi is a integrated SCSI and ATAPI layer under NetBSD and
	 * possibly also under other operatingsystems.
	 */
#include <sys/scsiio.h>
#include <dev/scsipi/scsipi_all.h>


int scsi_open(struct udf_discinfo *disc) {
	struct stat stat;
 
	disc->fhandle = open(disc->dev_name, O_RDWR, 0); /* no create */
	if (disc->fhandle<0) {
		perror("Failure to open device or file");
		return ENODEV;
	};
 
	if (fstat(disc->fhandle, &stat) < 0) {
		perror("Can't stat device or file");
		scsi_close(disc);
		return ENODEV;
	};

	return 0;
}


int scsi_close(struct udf_discinfo * disc) {
	close(disc->fhandle);
	disc->fhandle = -1;

	return 0;
}


int scsi_command(struct udf_discinfo *disc, void *cmd, size_t cmdlen, void *data, size_t datalen,  uint32_t timeout, int flags) {
	scsireq_t req;

	memset(&req, 0, sizeof(req));

	memcpy(req.cmd, cmd, cmdlen);
	req.cmdlen = cmdlen;
	req.databuf = data;
	req.datalen = datalen;
	req.timeout = timeout;
	req.flags = flags;
	req.senselen = SENSEBUFLEN;

	if (ioctl(disc->fhandle, SCIOCCOMMAND, &req) == -1)
		err(1, "SCIOCCOMMAND");

	if (req.retsts == SCCMD_OK)
		return 0;

	/* Some problem; report it and exit. */
	if (req.retsts == SCCMD_TIMEOUT) {
		if (scsilib_verbose)
			fprintf(stderr, "%s: SCSI command timed out\n", disc->dev_name);
		return EAGAIN;
	} else if (req.retsts == SCCMD_BUSY) {
		if (scsilib_verbose)
			fprintf(stderr, "%s: device is busy\n", disc->dev_name);
		return EBUSY;
	} else if (req.retsts == SCCMD_SENSE) {
		if (scsilib_verbose)
			scsi_print_sense((char *) disc->dev_name, req.cmd, req.cmdlen, req.sense, req.senselen_used, 1);
		return EIO;
	} else
		if (scsilib_verbose)
			fprintf(stderr, "%s: device had unknown status %x\n", disc->dev_name,
		  	  req.retsts);

	return EFAULT;
}


int scsi_check_for_scsi(struct udf_discinfo *disc) {
	struct scsi_addr	saddr;

	return ioctl(disc->fhandle, SCIOCIDENTIFY, &saddr);
}


int scsi_identify(struct udf_discinfo *disc, struct scsi_addr *saddr) {
	return ioctl(disc->fhandle, SCIOCIDENTIFY, saddr);
}
#endif	/* UDF_SCSIPI */



#ifdef UDF_LINUX_SCSI
	/*
	 * Support code for Linux SCSI code. It uses the ioctl() way of
	 * communicating since this is more close to the origional NetBSD
	 * scsipi implementation.
	 */
#include <scsi/sg.h>
#include <scsi/scsi.h>

#define SENSEBUFLEN 48


int scsi_open(struct udf_discinfo * disc) {
	int flags;
	struct stat stat;

	/* in Linux we are NOT allowed to open it blocking */
	disc->fhandle = open(disc->dev_name, O_RDWR | O_NONBLOCK, 0); /* no create */
	if (disc->fhandle<0) {
		perror("Failure to open device or file");
		return ENODEV;
	};

	/* explicitly mark it non blocking (again) (silly Linux) */
	flags = fcntl(disc->fhandle, F_GETFL);
	flags &= ~O_NONBLOCK;
	fcntl(disc->fhandle, F_SETFL, flags);

	if (fstat(disc->fhandle, &stat) < 0) {
		perror("Can't stat device or file");
		scsi_close(disc);
		return ENODEV;
	};

	return 0;
}


int scsi_close(struct udf_discinfo * disc) { 
	close(disc->fhandle);
	disc->fhandle = -1;

	return 0;
}


int scsi_command(struct udf_discinfo *disc, void *cmd, size_t cmdlen, void *data, size_t datalen,  uint32_t timeout, int flags) {
	struct sg_io_hdr req;
	uint8_t sense_buffer[SENSEBUFLEN];
	int error;

	bzero(&req, sizeof(req));
	if (flags == SG_DXFER_FROM_DEV) bzero(data, datalen);

	req.interface_id    = 'S';
	req.dxfer_direction = flags;
	req.cmd_len	    = cmdlen;
	req.mx_sb_len	    = SENSEBUFLEN;
	req.iovec_count	    = 0;
	req.dxfer_len	    = datalen;
	req.dxferp	    = data;
	req.cmdp	    = cmd;
	req.sbp		    = sense_buffer;
	req.flags	    = 0;
	req.timeout	    = timeout;

	error = ioctl(disc->fhandle, SG_IO, &req);

	if (req.status) {
		/* Is this OK? */
		if (scsilib_verbose) {
			scsi_print_sense((char *) disc->dev_name, cmd, cmdlen, sense_buffer, req.sb_len_wr, 1);
		};
	};

	return error;
}


int scsi_check_for_scsi(struct udf_discinfo *disc) {
	struct scsi_addr saddr;

	return scsi_identify(disc, &saddr);
}


int scsi_identify(struct udf_discinfo *disc, struct scsi_addr *saddr) {
	struct sg_scsi_id sg_scsi_id;
	struct sg_id {
		uint32_t tlci;	/* target | lun << 8 | channel << 16 | low_ino << 24 */
		uint32_t uniq_id;
	} sg_id;
	int emulated;
	int error;

	/* clean result */
	bzero(saddr, sizeof(struct scsi_addr));

	/* check if its really SCSI or emulated SCSI (ATAPI f.e.) */
	saddr->type = TYPE_SCSI;
	ioctl(disc->fhandle, SG_EMULATED_HOST, &emulated);
	if (emulated) saddr->type = TYPE_ATAPI;

	/* try 2.4 kernel or older */
	error = ioctl(disc->fhandle, SG_GET_SCSI_ID, &sg_scsi_id);
	if (!error) {
		saddr->addr.scsi.target = sg_scsi_id.scsi_id;
		saddr->addr.scsi.lun    = sg_scsi_id.lun;
		saddr->addr.scsi.scbus  = sg_scsi_id.channel;

		return 0;
	};

	/* 2.6 kernel or newer */
 	error = ioctl(disc->fhandle, SCSI_IOCTL_GET_IDLUN, &sg_id);
	if (error) return error;

	saddr->addr.scsi.target = (sg_id.tlci      ) & 0xff;
	saddr->addr.scsi.lun    = (sg_id.tlci >>  8) & 0xff;
	saddr->addr.scsi.scbus  = (sg_id.tlci >> 16) & 0xff;

	return error;
}

#endif	/* UDF_LINUX_SCSI */



#ifdef UDF_FREEBSD_CAM

int scsi_open(struct udf_discinfo * disc) {
	disc->devhandle = cam_open_device(disc->dev_name, O_RDWR);
	if (disc->devhandle == NULL) {
		disc->fhandle = open(disc->dev_name, O_RDWR | O_NONBLOCK, 0);
		if (disc->fhandle < 0) {
			perror("Failure to open device or file");
			return ENODEV;
		}
	}

	return 0;
}


int scsi_close(struct udf_discinfo * disc) {
	if(disc->devhandle != NULL) {
		cam_close_device(disc->devhandle);
		disc->devhandle = NULL;
	} else {
		close(disc->fhandle);
		disc->fhandle = -1;
	}

	return 0;
}


int scsi_command(struct udf_discinfo *disc, void *cmd, size_t cmdlen, void *data, size_t datalen,  uint32_t timeout, int flags) {
	struct cam_device *cam_dev;
	union ccb ccb;
	uint32_t cam_sense;

	memset(&ccb, 0, sizeof(ccb));
	cam_dev = (struct cam_device *) disc->devhandle;

	if (datalen == 0) flags = SCSI_NODATACMD;
	/* optional : */
	/* if (data) assert(flags == SCSI_NODATACMD); */

	cam_fill_csio(
		&ccb.csio,
		0,			/* retries */
		NULL,			/* cbfcnp */
		(flags & SCSI_READCMD) ? CAM_DIR_IN : ((flags & SCSI_WRITECMD) ? CAM_DIR_OUT : CAM_DIR_NONE),
					/* flags */
		MSG_SIMPLE_Q_TAG,	/* tag_action */
		(u_int8_t *) data,	/* data_ptr */
		datalen,		/* dxfer_len */
		SSD_FULL_SIZE,		/* sense_len */
		cmdlen,			/* cdb_len */
		timeout			/* timeout */
	);
 
	/* Disable freezing the device queue */
	ccb.ccb_h.flags |= CAM_DEV_QFRZDIS;
 
	memcpy(ccb.csio.cdb_io.cdb_bytes, cmd, cmdlen);

	/* Send the command down via the CAM interface */
	if (cam_send_ccb(cam_dev, &ccb) < 0) {
		err(1, "cam_send_ccb");
	}

	if ((ccb.ccb_h.status & CAM_STATUS_MASK) != CAM_REQ_CMP) {
		if (scsilib_verbose) {
			fprintf(stderr, "%s %s: %s=0x%x\n",
				"SCSI command error on", disc->dev_name,
				"Status", ccb.ccb_h.status);
		}

		cam_sense = (ccb.ccb_h.status & (CAM_STATUS_MASK | CAM_AUTOSNS_VALID));
		if (cam_sense == (CAM_SCSI_STATUS_ERROR | CAM_AUTOSNS_VALID)) {
			if (scsilib_verbose) {
				fprintf(stderr,
					"Sense data: Key=0x%x () "
					"Code=0x%x Qual=0x%x\n",
					ccb.csio.sense_data.flags & SSD_KEY,
					ccb.csio.sense_data.add_sense_code,
					ccb.csio.sense_data.add_sense_code_qual
				);
			}
		}
		return EFAULT;
	}

	return 0;
}



int scsi_identify(struct udf_discinfo * disc, struct scsi_addr *saddr) {
	struct cam_device * cam_dev = (struct cam_device *)disc->devhandle;

	/* clean result */
	bzero(saddr, sizeof(struct scsi_addr));

	if (!cam_dev) return ENODEV;

	/* check if its really SCSI or emulated SCSI (ATAPI f.e.) */
	saddr->type = TYPE_SCSI;
	saddr->addr.scsi.target = cam_dev->target_id;
	saddr->addr.scsi.lun    = cam_dev->target_lun;
	saddr->addr.scsi.scbus  = cam_dev->bus_id;

	return 0;
}


int scsi_check_for_scsi(struct udf_discinfo * disc) {
	struct scsi_addr saddr;

	return scsi_identify(disc, &saddr);
}

#endif	/* UDF_FREEBSD_CAM */


#ifndef UDF_FREEBSD_CAM
/*
 * Generic SCSI funtions also used by the sense printing functionality.
 * FreeBSD support has it allready asked for by the CAM.
 */

int scsi_mode_sense(struct udf_discinfo *disc, uint8_t pgcode, uint8_t pctl, void *buf, size_t len) {
	scsicmd cmd;

	bzero(buf, len);		/* initialise recieving buffer	*/

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x1a;			/* MODE SENSE			*/
	cmd[ 1] = 0;			/* -				*/
	cmd[ 2] = pgcode | pctl;	/* page code and control flags	*/
	cmd[ 3] = 0;			/* -				*/
	cmd[ 4] = len;			/* length of recieve buffer	*/
	cmd[ 5] = 0;			/* control			*/

	return scsi_command(disc, &cmd, 6, buf, len, 10000, SCSI_READCMD);
}


int scsi_mode_select(struct udf_discinfo *disc, uint8_t byte2, void *buf, size_t len) {
	scsicmd cmd;

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x15;			/* MODE SELECT			*/
	cmd[ 1] = 0x10 | byte2;		/* SCSI-2 page format select	*/
	cmd[ 4] = len;			/* length of page settings	*/
	cmd[ 5] = 0;			/* control			*/

	return scsi_command(disc, &cmd, 6, buf, len, 10000, SCSI_WRITECMD);
}


int scsi_request_sense(struct udf_discinfo *disc, void *buf, size_t len) {
	scsicmd cmd;

	bzero(buf, len);		/* initialise recieving buffer	*/

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x03;			/* REQUEST SENSE		*/
	cmd[ 4] = len;			/* length of data to be read	*/
	cmd[ 5] = 0;			/* control			*/

	return scsi_command(disc, &cmd, 6, buf, len, 10000, SCSI_WRITECMD);
}

#endif	/* !UDF_FREEBSD_CAM */


/* end of scsi_subr.c */

==> UDFclient.0.4.9.tgz: UDFclient.0.4.9/scsi_subr.c <==
/*	$NetBSD: scsi_subr.c,v 1.7 2002/10/08 20:17:06 soren Exp $	*/

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Charles M. Hannum; Jason R. Thorpe of the Numerical Aerospace
 * Simulation Facility, NASA Ames Research Center.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Small changes, generalisations and Linux support by Reinoud Zandijk
 * <reinoud@netbsd.org>.
 *
 */


/*
 * SCSI support subroutines.
 */

#include <sys/param.h>
#include <sys/ioctl.h>
#include <err.h> 
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <inttypes.h>
#include <assert.h>

#include "scsi_sense.h"
#include "scsilib.h"


#define	STRVIS_ISWHITE(x) ((x) == ' ' || (x) == '\0' || (x) == (u_char)'\377')
int scsilib_verbose = 0;


#ifdef UDF_SCSIPI
	/*
	 * scsipi is a integrated SCSI and ATAPI layer under NetBSD and exists
	 * in a modified form under OpenBSD and possibly also under other
	 * operatingsystems.
	 */

#include <sys/scsiio.h>
#ifdef __OpenBSD__
#include <scsi/scsi_all.h>
#else
#include <dev/scsipi/scsipi_all.h>
#endif


int scsi_open(struct udf_discinfo *disc) {
	struct stat stat;
 
	disc->fhandle = open(disc->dev_name, O_RDWR, 0); /* no create */
	if (disc->fhandle<0) {
		perror("Failure to open device or file");
		return ENODEV;
	};
 
	if (fstat(disc->fhandle, &stat) < 0) {
		perror("Can't stat device or file");
		scsi_close(disc);
		return ENODEV;
	};

	return 0;
}


int scsi_close(struct udf_discinfo * disc) {
	close(disc->fhandle);
	disc->fhandle = -1;

	return 0;
}


int scsi_command(struct udf_discinfo *disc, void *cmd, size_t cmdlen, void *data, size_t datalen,  uint32_t timeout, int flags) {
	scsireq_t req;

	memset(&req, 0, sizeof(req));

	memcpy(req.cmd, cmd, cmdlen);
	req.cmdlen = cmdlen;
	req.databuf = data;
	req.datalen = datalen;
	req.timeout = timeout;
	req.flags = flags;
	req.senselen = SENSEBUFLEN;

	if (ioctl(disc->fhandle, SCIOCCOMMAND, &req) == -1)
		err(1, "SCIOCCOMMAND");

	if (req.retsts == SCCMD_OK)
		return 0;

	/* Some problem; report it and exit. */
	if (req.retsts == SCCMD_TIMEOUT) {
		if (scsilib_verbose)
			fprintf(stderr, "%s: SCSI command timed out\n", disc->dev_name);
		return EAGAIN;
	} else if (req.retsts == SCCMD_BUSY) {
		if (scsilib_verbose)
			fprintf(stderr, "%s: device is busy\n", disc->dev_name);
		return EBUSY;
	} else if (req.retsts == SCCMD_SENSE) {
		if (scsilib_verbose)
			scsi_print_sense((char *) disc->dev_name, req.cmd, req.cmdlen, req.sense, req.senselen_used, 1);
		return EIO;
	} else
		if (scsilib_verbose)
			fprintf(stderr, "%s: device had unknown status %x\n", disc->dev_name,
		  	  req.retsts);

	return EFAULT;
}


/*
 * The reasoning behind this explicit copy is for compatibility with changes
 * in our uscsi_addr structure.
 */
int scsi_identify(struct udf_discinfo *disc, struct uscsi_addr *saddr) {
	struct scsi_addr raddr;
	int error;

	bzero(saddr, sizeof(struct uscsi_addr));
	error = ioctl(disc->fhandle, SCIOCIDENTIFY, &raddr);
	if (error) return error;

#ifdef __NetBSD__
	/* scsi and atapi are split up like in uscsi_addr */
	if (raddr.type == 0) {
		saddr->type = USCSI_TYPE_SCSI;
		saddr->addr.scsi.scbus  = raddr.addr.scsi.scbus;
		saddr->addr.scsi.target = raddr.addr.scsi.target;
		saddr->addr.scsi.lun    = raddr.addr.scsi.lun;
	} else {
		saddr->type = USCSI_TYPE_ATAPI;
		saddr->addr.atapi.atbus = raddr.addr.atapi.atbus;
		saddr->addr.atapi.drive = raddr.addr.atapi.drive;
	};
#endif
#ifdef __OpenBSD__
	/* atapi's are shown as SCSI devices */
	if (raddr.type == 0) {
		saddr->type = USCSI_TYPE_SCSI;
		saddr->addr.scsi.scbus  = raddr.scbus;
		saddr->addr.scsi.target = raddr.target;
		saddr->addr.scsi.lun    = raddr.lun;
	} else {
		saddr->type = USCSI_TYPE_ATAPI;
		saddr->addr.atapi.atbus = raddr.scbus;	/* overload */
		saddr->addr.atapi.drive = raddr.target;	/* overload */
	};
#endif

	return 0;
}


int scsi_check_for_scsi(struct udf_discinfo *disc) {
	struct uscsi_addr	saddr;

	return scsi_identify(disc, &saddr);
}
#endif	/* UDF_SCSIPI */







#ifdef UDF_LINUX_SCSI
	/*
	 * Support code for Linux SCSI code. It uses the ioctl() way of
	 * communicating since this is more close to the origional NetBSD
	 * scsipi implementation.
	 */
#include <scsi/sg.h>
#include <scsi/scsi.h>

#define SENSEBUFLEN 48


int scsi_open(struct udf_discinfo * disc) {
	int flags;
	struct stat stat;

	/* in Linux we are NOT allowed to open it blocking */
	disc->fhandle = open(disc->dev_name, O_RDWR | O_NONBLOCK, 0); /* no create */
	if (disc->fhandle<0) {
		perror("Failure to open device or file");
		return ENODEV;
	};

	/* explicitly mark it non blocking (again) (silly Linux) */
	flags = fcntl(disc->fhandle, F_GETFL);
	flags &= ~O_NONBLOCK;
	fcntl(disc->fhandle, F_SETFL, flags);

	if (fstat(disc->fhandle, &stat) < 0) {
		perror("Can't stat device or file");
		scsi_close(disc);
		return ENODEV;
	};

	return 0;
}


int scsi_close(struct udf_discinfo * disc) { 
	close(disc->fhandle);
	disc->fhandle = -1;

	return 0;
}


int scsi_command(struct udf_discinfo *disc, void *cmd, size_t cmdlen, void *data, size_t datalen,  uint32_t timeout, int flags) {
	struct sg_io_hdr req;
	uint8_t sense_buffer[SENSEBUFLEN];
	int error;

	bzero(&req, sizeof(req));
	if (flags == SG_DXFER_FROM_DEV) bzero(data, datalen);

	req.interface_id    = 'S';
	req.dxfer_direction = flags;
	req.cmd_len	    = cmdlen;
	req.mx_sb_len	    = SENSEBUFLEN;
	req.iovec_count	    = 0;
	req.dxfer_len	    = datalen;
	req.dxferp	    = data;
	req.cmdp	    = cmd;
	req.sbp		    = sense_buffer;
	req.flags	    = 0;
	req.timeout	    = timeout;

	error = ioctl(disc->fhandle, SG_IO, &req);

	if (req.status) {
		/* Is this OK? */
		if (scsilib_verbose) {
			scsi_print_sense((char *) disc->dev_name, cmd, cmdlen, sense_buffer, req.sb_len_wr, 1);
		};
	};

	return error;
}


int scsi_identify(struct udf_discinfo *disc, struct uscsi_addr *saddr) {
	struct sg_scsi_id sg_scsi_id;
	struct sg_id {
		uint32_t tlci;	/* target | lun << 8 | channel << 16 | low_ino << 24 */
		uint32_t uniq_id;
	} sg_id;
	int emulated;
	int error;

	/* clean result */
	bzero(saddr, sizeof(struct uscsi_addr));

	/* check if its really SCSI or emulated SCSI (ATAPI f.e.) */
	saddr->type = USCSI_TYPE_SCSI;
	ioctl(disc->fhandle, SG_EMULATED_HOST, &emulated);
	if (emulated) saddr->type = USCSI_TYPE_ATAPI;

	/* try 2.4 kernel or older */
	error = ioctl(disc->fhandle, SG_GET_SCSI_ID, &sg_scsi_id);
	if (!error) {
		saddr->addr.scsi.target = sg_scsi_id.scsi_id;
		saddr->addr.scsi.lun    = sg_scsi_id.lun;
		saddr->addr.scsi.scbus  = sg_scsi_id.channel;

		return 0;
	};

	/* 2.6 kernel or newer */
 	error = ioctl(disc->fhandle, SCSI_IOCTL_GET_IDLUN, &sg_id);
	if (error) return error;

	saddr->addr.scsi.target = (sg_id.tlci      ) & 0xff;
	saddr->addr.scsi.lun    = (sg_id.tlci >>  8) & 0xff;
	saddr->addr.scsi.scbus  = (sg_id.tlci >> 16) & 0xff;

	return error;
}


int scsi_check_for_scsi(struct udf_discinfo *disc) {
	struct uscsi_addr saddr;

	return scsi_identify(disc, &saddr);
}
#endif	/* UDF_LINUX_SCSI */







#ifdef UDF_FREEBSD_CAM

int scsi_open(struct udf_discinfo *disc) {
	disc->devhandle = cam_open_device(disc->dev_name, O_RDWR);

	if (disc->devhandle == NULL) {
		disc->fhandle = open(disc->dev_name, O_RDWR | O_NONBLOCK, 0);
		if (disc->fhandle < 0) {
			perror("Failure to open device or file");
			return ENODEV;
		}
	}

	return 0;
}


int scsi_close(struct udf_discinfo *disc) {
	if (disc->devhandle != NULL) {
		cam_close_device(disc->devhandle);
		disc->devhandle = NULL;
	} else {
		close(disc->fhandle);
		disc->fhandle = -1;
	}

	return 0;
}


int scsi_command(struct udf_discinfo *disc, void *cmd, size_t cmdlen, void *data, size_t datalen,  uint32_t timeout, int flags) {
	struct cam_device *cam_dev;
	union ccb ccb;
	uint32_t cam_sense;

	memset(&ccb, 0, sizeof(ccb));
	cam_dev = (struct cam_device *) disc->devhandle;

	if (datalen == 0) flags = SCSI_NODATACMD;
	/* optional : */
	/* if (data) assert(flags == SCSI_NODATACMD); */

	cam_fill_csio(
		&ccb.csio,
		0,			/* retries */
		NULL,			/* cbfcnp */
		(flags & SCSI_READCMD) ? CAM_DIR_IN : ((flags & SCSI_WRITECMD) ? CAM_DIR_OUT : CAM_DIR_NONE),
					/* flags */
		MSG_SIMPLE_Q_TAG,	/* tag_action */
		(u_int8_t *) data,	/* data_ptr */
		datalen,		/* dxfer_len */
		SSD_FULL_SIZE,		/* sense_len */
		cmdlen,			/* cdb_len */
		timeout			/* timeout */
	);
 
	/* Disable freezing the device queue */
	ccb.ccb_h.flags |= CAM_DEV_QFRZDIS;
 
	memcpy(ccb.csio.cdb_io.cdb_bytes, cmd, cmdlen);

	/* Send the command down via the CAM interface */
	if (cam_send_ccb(cam_dev, &ccb) < 0) {
		err(1, "cam_send_ccb");
	}

	if ((ccb.ccb_h.status & CAM_STATUS_MASK) != CAM_REQ_CMP) {
		if (scsilib_verbose) {
			fprintf(stderr, "%s %s: %s=0x%x\n",
				"SCSI command error on", disc->dev_name,
				"Status", ccb.ccb_h.status);
		}

		cam_sense = (ccb.ccb_h.status & (CAM_STATUS_MASK | CAM_AUTOSNS_VALID));
		if (cam_sense == (CAM_SCSI_STATUS_ERROR | CAM_AUTOSNS_VALID)) {
			if (scsilib_verbose) {
				/* maybe print using the scsi_sense routines? */
				fprintf(stderr,
					"Sense data: Key=0x%x () "
					"Code=0x%x Qual=0x%x\n",
					ccb.csio.sense_data.flags & SSD_KEY,
					ccb.csio.sense_data.add_sense_code,
					ccb.csio.sense_data.add_sense_code_qual
				);
			}
		}
		return EFAULT;
	}

	return 0;
}



int scsi_identify(struct udf_discinfo *disc, struct uscsi_addr *saddr) {
	struct cam_device *cam_dev;
	
	/* clean result */
	bzero(saddr, sizeof(struct uscsi_addr));

	cam_dev = (struct cam_device *) disc->devhandle;
	if (!cam_dev) return ENODEV;

	/* check if its really SCSI or emulated SCSI (ATAPI f.e.) ? */
	saddr->type = USCSI_TYPE_SCSI;
	saddr->addr.scsi.target = cam_dev->target_id;
	saddr->addr.scsi.lun    = cam_dev->target_lun;
	saddr->addr.scsi.scbus  = cam_dev->bus_id;

	return 0;
}


int scsi_check_for_scsi(struct udf_discinfo *disc) {
	struct uscsi_addr saddr;

	return scsi_identify(disc, &saddr);
}

#endif	/* UDF_FREEBSD_CAM */






#ifndef UDF_FREEBSD_CAM
/*
 * Generic SCSI funtions also used by the sense printing functionality.
 * FreeBSD support has it allready asked for by the CAM.
 */

int scsi_mode_sense(struct udf_discinfo *disc, uint8_t pgcode, uint8_t pctl, void *buf, size_t len) {
	scsicmd cmd;

	bzero(buf, len);		/* initialise recieving buffer	*/

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x1a;			/* MODE SENSE			*/
	cmd[ 1] = 0;			/* -				*/
	cmd[ 2] = pgcode | pctl;	/* page code and control flags	*/
	cmd[ 3] = 0;			/* -				*/
	cmd[ 4] = len;			/* length of recieve buffer	*/
	cmd[ 5] = 0;			/* control			*/

	return scsi_command(disc, &cmd, 6, buf, len, 10000, SCSI_READCMD);
}


int scsi_mode_select(struct udf_discinfo *disc, uint8_t byte2, void *buf, size_t len) {
	scsicmd cmd;

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x15;			/* MODE SELECT			*/
	cmd[ 1] = 0x10 | byte2;		/* SCSI-2 page format select	*/
	cmd[ 4] = len;			/* length of page settings	*/
	cmd[ 5] = 0;			/* control			*/

	return scsi_command(disc, &cmd, 6, buf, len, 10000, SCSI_WRITECMD);
}


int scsi_request_sense(struct udf_discinfo *disc, void *buf, size_t len) {
	scsicmd cmd;

	bzero(buf, len);		/* initialise recieving buffer	*/

	bzero(cmd, SCSI_CMD_LEN);
	cmd[ 0] = 0x03;			/* REQUEST SENSE		*/
	cmd[ 4] = len;			/* length of data to be read	*/
	cmd[ 5] = 0;			/* control			*/

	return scsi_command(disc, &cmd, 6, buf, len, 10000, SCSI_WRITECMD);
}

#endif	/* !UDF_FREEBSD_CAM */


/* end of scsi_subr.c */
